/* file KUT/kutreq.c - requests */
/* Time-stamp: <1998 Augu 30 Sun 11:57:33 CEST Basile.Starynkevitch@wanadoo.fr> */
/* rcsid: $Id: kutreq.c,v 0.4 1998/09/02 20:29:50 basile Exp $ */

#include "kut.h"

const char kutreq_ident[] =
 "@(#)$Id: kutreq.c,v 0.4 1998/09/02 20:29:50 basile Exp $"
 " <" __DATE__ "@" __TIME__ ">" ;


enum reqstate_en { rqs_none=0, rqs_pending, rqs_ended };

/* a request has a unique id - it is created inside facilities and can
   be waited for */
struct request_st {
  /* unique request id */
  reqid_t rq_id;
  /* request type */
  unsigned char rq_type;
  /* request state */
  unsigned char rq_state;
  /* rank of next request in same bucket modulo REQUEST_BUCKET_NB */
  unsigned short rq_next;		 
  /* request data and private - for facility */
  void* rq_data;
  unsigned rq_priv;
  /* request result - returned to user */
  unsigned rq_res;
};

/* the request table can be seen as an hashed hashed table:
   - requests are put in reqtable after rank rqid%REQUEST_TABLE_LEN 
     but never at index 0
   - also requests are hashbucked thru reqbucket and rq_next */

#define REQUEST_TABLE_LEN 197 /* should be prime and >100 and <32767! */
#define REQUEST_BUCKET_NB 64 /* a power of 2 a few times less than length */
static struct request_st reqtable[REQUEST_TABLE_LEN];
static unsigned short reqbucket[REQUEST_BUCKET_NB];

static int nb_requests=0;
static int ridseed= 12175;

reqid_t request_create(unsigned char typ, void*data, unsigned priv) {
  int intlev;
  reqid_t rqid = 0;
  unsigned rk;
  int cnt;
  unsigned br;
  intlev = int_saved();
  /* request table almost full? */
  if (nb_requests>(4*REQUEST_TABLE_LEN)/5) { rqid=0; goto end;};
  ridseed += (typ&0x1f);
  do { /* often loop done once */
    ridseed= ridseed*3020491+28619; /* both are prime */
    ridseed &= 0x3fffffff;
    ridseed ++; /* so it is always >0! */
    for ( (cnt=REQUEST_TABLE_LEN/4),
	  (rk= ridseed % REQUEST_TABLE_LEN);
	  cnt>0;
	  cnt--) {
      if (rk==0) rk++;
      if (reqtable[rk].rq_id == 0) {
	/* common case - found empty slot */
	cnt=0; 
	rqid=ridseed;
	break;
      }
      else if (reqtable[rk].rq_id == ridseed) {
	/* rare case: duplicate ridseed, so try another one! */
	cnt=0;
	rqid=0;
	break;
      } else if (rk<REQUEST_TABLE_LEN)
	rk++;
      else rk=1;
    };
    if (rqid!=0) {
      /* checkout that rqid is not duplicate thru the bucket list */
      for (br=reqbucket[rqid%REQUEST_BUCKET_NB];
	   br>0;
	   br=reqtable[br].rq_next)
	if (reqtable[br].rq_id == ridseed) {
	  rqid=0;
	  break;
	};
    };
  } while (rqid==0);
  /* here rqid is >0, unique, and should go at rank rk>0 */
  reqtable[rk].rq_id = rqid;
  reqtable[rk].rq_type = typ;
  reqtable[rk].rq_data = data;
  reqtable[rk].rq_priv = priv;
  reqtable[rk].rq_next = reqbucket[rqid%REQUEST_BUCKET_NB];
  reqtable[rk].rq_res = 0;
  reqtable[rk].rq_state = rqs_pending;
  reqbucket[rqid%REQUEST_BUCKET_NB] = rk;
  nb_requests ++;
end: 
  int_restore(intlev);
  return rqid;
}/* end of request_create */


/* internal routine to find a request address by its id */
static inline struct request_st* request_ptr(reqid_t rqid) {
  unsigned rk;
  if (rqid<=0) return (struct request_st*)0;
  rk = rqid%REQUEST_TABLE_LEN;
  if (rk<REQUEST_TABLE_LEN-4) {
    /* heuristically happens often and handle rk==0 case */
    if (reqtable[rk].rq_id == rqid) return &(reqtable[rk]);
    rk++;
    if (reqtable[rk].rq_id == rqid) return &(reqtable[rk]);
    rk++;
    if (reqtable[rk].rq_id == rqid) return &(reqtable[rk]);
    rk++;
    if (reqtable[rk].rq_id == rqid) return &(reqtable[rk]);
  };
  /* systematic search thru bucket */
  for (rk=reqbucket[rqid%REQUEST_BUCKET_NB]; rk>0; 
       rk=reqtable[rk].rq_next) 
    if (reqtable[rk].rq_id == rqid) return &(reqtable[rk]);
  return (struct request_st*)0;
}

static reqid_t ended_rqid=0;

/* called with disabled interrupts from facility or interrupt handler
   - return 0 iff ok */
int request_end(reqid_t rqid, unsigned res) {
  struct request_st *rq = request_ptr(rqid);
  if (!rq) return 1;
  rq->rq_res = res;
  rq->rq_state = rqs_ended;  
  ended_rqid = rqid;
#ifdef KUT_DEBUG
  scrs_si("request_end rqid=", rqid);
  scrs_sin(" res=", res);
  {int i; for (i=100000; i>0; i--) asm volatile ("nop");};
#endif /*KUT_DEBUG*/
  return 0;
}

/* called with disabled interrupts */
static inline void remove_request_ptr(struct request_st* rq)
{
  reqid_t rqid = rq->rq_id;
  unsigned short rk;
  unsigned short *prec;
  prec= &(reqbucket[rqid%REQUEST_BUCKET_NB]);
  for (rk=reqbucket[rqid%REQUEST_BUCKET_NB]; rk>0; 
       rk=reqtable[rk].rq_next) {
    if (reqtable[rk].rq_id == rqid) {
      *prec = rq->rq_next;
      break;
    };
    prec = &(reqtable[rk].rq_next);
  };
  rq->rq_id = 0;
  rq->rq_type = 0;
  rq->rq_data = 0;
  rq->rq_priv = 0;
  rq->rq_next = 0;
  rq->rq_res = 0;
  rq->rq_state = rqs_none;
  if (ended_rqid == rqid) ended_rqid=0;
  nb_requests --;
}

/***** waiting primitives - called by usercode with interrupts enabled *****/

#define halt() asm volatile ("hlt" : : : "memory")

/* wait a given request - return 0 iff ok and set *pres to result */
int request_wait(reqid_t rqid, unsigned *pres) {
  int intlev= int_saved();
  struct request_st *rq = request_ptr(rqid);
  if (!rq) goto errend;
  if (rq->rq_state == rqs_none) goto errend;
  while (rq->rq_state == rqs_pending) {
    sti();
    asm volatile(".globl _halt_request_wait; _halt_request_wait: hlt; nop");
    /* interrupts occur here! */
    cli();
  }/* end while */;
  if (rq->rq_state == rqs_ended) {
    if (pres) *pres=rq->rq_res;
    remove_request_ptr(rq);
    goto end;
  } else 
    goto errend;
 end:
  int_restore(intlev);
  return 0;
 errend:
  int_restore(intlev);
  return 1;
}

/* wait for one of a given array of requests */
reqid_t request_wait_some(reqid_t rqidtable[], int len, unsigned *pres) {
  struct request_st *rqtab[REQARRAY_MAXLEN];
  int intlev;
  reqid_t rqid=0;
  int i;
  if (len<=0 || len>=REQARRAY_MAXLEN) 
    return 0;
  intlev= int_saved();
  for (i=len-1; i>=0; i--) 
    if (!(rqtab[i] = request_ptr(rqidtable[i]))) 
      goto end;
  for (;;) {
    /* find an ended request */
    for (i=len-1; i>=0; i--) {
      if (rqtab[i]->rq_state == rqs_none) goto end;
      if (rqtab[i]->rq_state == rqs_ended) {
	rqid = rqtab[i]->rq_id;
	*pres = rqtab[i]->rq_res;
	remove_request_ptr(rqtab[i]);
	goto end;
      }
    };
    sti();
    asm volatile(".globl _halt_request_wait_some; _halt_request_wait_some: hlt; nop");
    /* interrupts occur here! */
    cli();
  } 
 end:
  int_restore(intlev);
  return rqid;
}

/* wait for all of a given array of requests - return 0 iff ok*/
int request_wait_all(reqid_t rqidtable[], unsigned restable[], int len) {
  struct request_st *rqtab[REQARRAY_MAXLEN];
  int intlev;
  int i;
  int err=1;
  int nbended=0;
  if (len<=0 || len>=REQARRAY_MAXLEN) 
    return 0;
  intlev= int_saved();
  for (i=len-1; i>=0; i--) 
    if (!(rqtab[i] = request_ptr(rqidtable[i]))) 
      goto end;
  for (;;) {
    /* find a pending request */
    for (i=len-1; i>=0; i--) {
      if (!rqtab[i]) 
	continue;
      if (rqtab[i]->rq_state == rqs_none) goto end;
      if (rqtab[i]->rq_state == rqs_ended) {
	restable[i] = rqtab[i]->rq_res;
	remove_request_ptr(rqtab[i]);
	nbended++;
	rqtab[i]=0;
      }
    };
    if (nbended>=len) {
      err=0;
      goto end;
    };
    sti();
    /* interrupts occur here! */
    asm volatile(".globl _halt_request_wait_all; _halt_request_wait_all: hlt; nop");
    cli();
  } 
 end:
  int_restore(intlev);
  return err;
}

/* wait for any arbitrary pending request */
reqid_t request_wait_any(unsigned *pres) {
  int intlev;
  int br, rk;
  reqid_t rqid=0;
  struct request_st *rq; 
#if 0 && KUT_DEBUG 
    scrs_sn("request_wait_any starting");
#endif /*KUT_DEBUG*/
  intlev= int_saved();
  for (;;) {
#if 0 && KUT_DEBUG
    scrs_sin("request_wait_any ended_rqid=", ended_rqid);
#endif /*KUT_DEBUG*/
    /* perhaps already got an ended request */
    if (ended_rqid>0) {
      rq = request_ptr(ended_rqid);
      if (rq->rq_state == rqs_ended) {
	if (pres) *pres = rq->rq_res;
	rqid=ended_rqid;
	remove_request_ptr(rq);
	goto end;
      }
    };
#if 0 && KUT_DEBUG
    scrs_sin("request_wait_any nb_requests=", nb_requests);
#endif /*KUT_DEBUG*/
    /* scan all requests */
    if (nb_requests >0) {
      for (br=0; br<REQUEST_BUCKET_NB; br++)
	for (rk=reqbucket[br]; rk>0; rk=reqtable[rk].rq_next) {
	  rq = reqtable+rk;
	  if (rq->rq_state == rqs_ended) {
	    if (pres) *pres = rq->rq_res;
	    rqid=rq->rq_id;
	    remove_request_ptr(rq);
	    goto end;
	  }
	}
    }
    /* pause for interrupts */
#if 0 && KUT_DEBUG
    scrs_sn("request_wait_any halting");
#endif /*KUT_DEBUG*/
    sti();
    asm volatile(".globl _halt_request_wait_any; _halt_request_wait_any: hlt; nop");
    /* interrupts occur here! */
    cli();
  }
 end:
  int_restore(intlev);
#ifdef KUT_DEBUG
    scrs_sin("request_wait_any rqid=", rqid);
#endif /*KUT_DEBUG*/
  return rqid;
}

/* test if request is completed without waiting */
int request_completed(reqid_t rqid, unsigned *pres) {
  int intlev= int_saved();
  struct request_st *rq = request_ptr(rqid);
  if (!rq) goto errend;
  if (rq->rq_state == rqs_ended) 
    goto end;
 errend:
  int_restore(intlev);
  return 0;
 end:
  int_restore(intlev);
  return 1;
}

/* ask for a request to be signaled asynchronously - return 0 iff ok */
int request_signal(reqid_t rqid, void*data, void (*handler)(reqid_t,void*)) {
  panic("request_signal unimplemented");
#warning unimplemented request_signal
  return 1;
}

/*** rcslog ***
 $Log: kutreq.c,v $
 Revision 0.4  1998/09/02 20:29:50  basile
 added labels in halting asm code (for debugging)

 Revision 0.3  1998/08/24 20:19:50  basile
 I'm not sure it is working

 Revision 0.2  1998/08/23 11:09:37  basile
 added ident string

 Revision 0.1  1998/08/22 17:47:32  basile
 coded most request related routine - but no handlers yet!

 Revision 0.0  1998/08/20 20:34:59  basile
 *** endlog ***/

/* eof $Header: /home/basile/KUT/RCS/kutreq.c,v 0.4 1998/09/02 20:29:50 basile Exp $ */
